
-- Exterior algebra

{-# LANGUAGE CPP, MultiParamTypeClasses, TypeFamilies, TypeSynonymInstances,
             GeneralizedNewtypeDeriving
  #-}
module Math.Algebra.PolyAlg.Exterior where

--------------------------------------------------------------------------------

import Control.Applicative ( (<$>) )

import Data.List
import Data.Maybe
import Data.Ord

import Text.ParserCombinators.Parsec

import qualified Math.Combinat.Permutations as Perm

import Data.FreeModule
import Math.Algebra.PolyAlg.Class

--------------------------------------------------------------------------------

{-
#define ASCII

#ifndef ASCII  
wedgeSymbol = "∧"  
#else
wedgeSymbol = "/\\"
#endif
-}

wedgeSymbol = "/\\\\"

--------------------------------------------------------------------------------
  
(<#>) :: (a -> b) -> (c -> d) -> (a,c) -> (b,d)
(f<#>g) (x,y) = (f x, g y)

--------------------------------------------------------------------------------

newtype ExteriorMonomial symb = EM [symb] deriving (Eq,Ord,Show)

type ExteriorTerm symb coeff = (ExteriorMonomial symb, coeff)

-- | The result may be zero if there is a repeated symbol; this case is encoded
-- as 'Nothing'.
toExteriorMonomial :: (Ord symb, Num coeff) => [symb] -> Maybe (ExteriorTerm symb coeff)
toExteriorMonomial xs = if b then Just (EM ys, sign) else Nothing where
  ps = sortBy (comparing snd) (zip [1..] xs)
  ys = map snd ps
  b  = and [ 1 == length z | z <- group ys ] 
  sign = case ps of
    [] -> 1
    _  -> (Perm.signOfPermutation . Perm.toPermutationUnsafe) (map fst ps)
  
toExteriorTerm :: (Ord symb, Num coeff) => [symb] -> coeff -> Maybe (ExteriorTerm symb coeff)
toExteriorTerm xs c = (id<#>(*c)) <$> (toExteriorMonomial xs)

mulExteriorTerm 
  :: (Ord symb, Num coeff) 
  => ExteriorTerm symb coeff -> ExteriorTerm symb coeff
  -> Maybe (ExteriorTerm symb coeff)
mulExteriorTerm (EM xs, c) (EM ys, d) = toExteriorTerm (xs++ys) (c*d)
  
showExteriorMonomialWith :: (symb -> String) -> ExteriorMonomial symb -> String
showExteriorMonomialWith showSymb (EM xs) 
  = concat $ intersperse wedgeSymbol $ map showSymb xs

instance Pretty symb => Pretty (ExteriorMonomial symb) where
  pretty = showExteriorMonomialWith pretty

--------------------------------------------------------------------------------

exteriorMonomP :: Num c => Parser (ExteriorTerm Symbol c)
exteriorMonomP = (f . toExteriorMonomial) <$> sepBy symbolP sep where
  sep = string "∧" <|> string "/\\" 
  f (Just x)  = x
  f (Nothing) = (EM [],0)

extAlgP :: Parser (ExtAlg Integer)
extAlgP = E <$> freeModuleP' exteriorMonomP integerP

parseExtAlg :: String -> ExtAlg Integer   
parseExtAlg s = case runParser extAlgP () "exterior \"polynomial\"" s of
  Left err -> error (show err)
  Right x  -> x 
    
--------------------------------------------------------------------------------

{-
newtype (FreeModule fmod, ExteriorMonomial symb ~ Base fmod, coeff ~ Coeff fmod) 
  => MultiVector fmod symb coeff = MV fmod
  
type ExtAlg c = MultiVector FreeMod Symbol c
-}

newtype ExtAlgebra symb coeff 
  = E { unE :: FreeMod (ExteriorMonomial symb) coeff } 
  deriving (Eq,Ord,Show,Pretty)

type ExtAlg c = ExtAlgebra Symbol c

-- "deriving FreeModule" does not work :(
instance (Ord s, Num c) => FreeModule (ExtAlgebra s c) where
  type Base  (ExtAlgebra s c) = ExteriorMonomial s
  type Coeff (ExtAlgebra s c) = c
  (E x) ^+^ (E y) = E (x ^+^ y)
  (E x) ^-^ (E y) = E (x ^-^ y)
  scalarMul c (E x) = E (scalarMul c x)
  fromList = E . fromList
  toList = toList . unE
  coeff b (E x) = coeff b x
  fromBase = E . fromBase
  fromTerm b c = E (fromTerm b c)
  zero = E zero
  size   = size . unE
  isZero = isZero . unE
  unionWith f (E x) (E y) = E (unionWith f x y)
  minTerm = minTerm . unE
  maxTerm = maxTerm . unE
  split = (E<#>E) . split . unE
  unsafeJoin (E x) (E y) = E (unsafeJoin x y)

(/\) :: (Ord s, Num c) => ExtAlgebra s c -> ExtAlgebra s c -> ExtAlgebra s c
(E a) /\ (E b) = 
  (E . fromList . catMaybes) 
    [ mulExteriorTerm x y | x <- toList a , y <- toList b ]

fromLinear :: (Ord s, Num c) => FreeMod s c -> ExtAlgebra s c 
fromLinear = baseMap (\x -> EM [x])

instance (Ord s, Num c) => Algebra (ExtAlgebra s c) where 
  (^*^) = (/\)
  
--------------------------------------------------------------------------------


